#! /usr/bin/env python
import os
import sys
import numpy as np
import matplotlib.pyplot as plt
from phoebe.backend import fitting
from phoebe.backend import universe
from phoebe.parameters import parameters

if __name__=="__main__":
    import sys
    if len(sys.argv[1:])==2:
        system = universe.load(sys.argv[1])
        fitparams = parameters.load(sys.argv[2])
        fitting.summarize_fit(system,fitparams)
    elif len(sys.argv[1:])==1 and os.path.basename(sys.argv[1]).split('.')[0]=='emcee_chain':
        emcee_chain = np.loadtxt(sys.argv[1]).T
        walkers = np.unique(emcee_chain[1])
        
        for i,ipar in enumerate(emcee_chain[2:]):
            plt.figure(i)
            plt.subplot(121)
            plt.gca().set_color_cycle([plt.cm.spectral(k) for k in np.linspace(0,1,len(walkers))])
            plt.ylabel("Parameter {}".format(i))
            for walker in walkers:
                keep = emcee_chain[1]==walker
                plt.plot(ipar[keep],'o-')
            plt.subplot(122)
            plt.hist(ipar,bins=50)
        
        for i,ipar in enumerate(emcee_chain[2:-1]):
            for j,jpar in enumerate(emcee_chain[2+i+1:]):
                plt.figure(100*i+j+100)
                plt.xlabel("Parameter {}".format(i))
                plt.ylabel("Parameter {}".format(i+1))
                plt.hexbin(ipar,jpar)
                
            
            
            
        
    
    plt.show()